unit uFileList;
{$INCLUDE 'Plugin.inc'}

interface

uses
  Windows, SysUtils, Classes,
  Generics.Collections, Generics.Defaults,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  uOptions,
  uFiles;

type
  TFileList = class(TFarEditorManaged)
  private
    fSourceDirectory: string;
    fOptions: TOptions;
    fRepeatedRun: boolean;
  protected
    procedure AddFileName(Text: TTextStream; const FileName: string; MaxFileNameLength: integer);
    function GetTranslatedFileName(const FileName: string): string;
    function DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionStart: integer): boolean; overload;
    function DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean; overload;
    function GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
    procedure BeforeClose(var CanClose: boolean; var Result: TIntPtr); override;
    procedure AfterReOpen; override;
    function ProcessKeyEvent(const Info: TKeyEventRecord): boolean; override;
    function GetDescription: string;
    type TInternalJumpResult = (ijFailed, ijSucceeded, ijAlreadyThere);
    function InternalJumpToDestinationFileName(out Row, CurPos: TIntPtr): TInternalJumpResult;
    function InternalJumpToDestinationFileExt(out Row, CurPos, LineLen: TIntPtr): TInternalJumpResult;
    property RepeatedRun: boolean read fRepeatedRun;
  public
    class function FindByEditorID(EditorID: TIntPtr; out Editor: TFileList): boolean;
    constructor Create(Options: TOptions);
    destructor Destroy; override;
    function LoadFromPanel(AActivePanel: boolean): boolean; overload;
    function LoadFromPanel(const AInfo: TFarPanel): boolean; overload;
    function LoadFromFileList(const Directory: string; const Files: TStrings): boolean;
    function OpenInEditor: boolean;
    function RenameFiles: boolean;
    function IsEditorLine: boolean;
    function SelectDestinationFilenames: boolean;
    function JumpToDestinationFileName: boolean;
    function JumpToDestinationFileExt: boolean;
    function SmartJumpToBegin: boolean;
    function SmartJumpToEnd: boolean;
    function DeleteTillExtension: boolean;
    function ShowEditorMenu: boolean;
    property SourceDirectory: string read fSourceDirectory write fSourceDirectory;
    property Options: TOptions read fOptions;
  end;

implementation

uses
  uSystem, uMessages, uPluginConsts;

{ TFileList }

procedure TFileList.AddFileName(Text: TTextStream; const FileName: string; MaxFileNameLength: integer);
begin
  Text.WriteLine(Format('%-*.*s    %s', [MaxFileNameLength, MaxFileNameLength, FileName, FileName]));
end;

procedure TFileList.AfterReOpen;
begin
  inherited;
  fRepeatedRun := True;
end;

procedure TFileList.BeforeClose(var CanClose: boolean; var Result: TIntPtr);
begin
  inherited;
  try
    CanClose := RenameFiles;
  except
    on EAbort do
      CanClose := True;
  end;
end;

constructor TFileList.Create(Options: TOptions);
begin
  inherited Create(SystemTempFile(SystemTempDir, 'FRename3', '.txt'));
  fOptions := Options;
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionStart: integer): boolean;
var
  i, n: integer;
begin
  Result := False;
  n := Length(Line);
  if n > 0 then
  begin
    if Line[1] = '"' then
    begin
      SourceFileNameStart := 2;
      i := 2;
      while (i <= n) and (Line[i] <> '"') do
        Inc(i);
      SourceFileNameLength := i - SourceFileNameStart;
      Inc(i);
      while (i <= n) and (Line[i] = ' ') do
        Inc(i);
      if (i <= n) and (Line[i] = '"') then
      begin
        Inc(i);
        DestinationFileNameStart := i;
        DestinationExtensionStart := -1;
        while (i <= n) and (Line[i] <> '"') do
        begin
          if Line[i] = '.' then
            DestinationExtensionStart := i;
          Inc(i);
        end;
        if DestinationExtensionStart < 0 then
          DestinationExtensionStart := i;
        DestinationFileNameLength := i - DestinationFileNameStart;
        Inc(i);
        while (i <= n) and (Line[i] = ' ') do
          Inc(i);
        if Pred(i) = n then
          Result := True;
      end;
    end;
  end;
end;

function TFileList.DecodeEditorLine(const Line: string; out SourceFileName, DestinationFileName: string): boolean;
var
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos: integer;
begin
  Result := False;
  if not DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos) then
    Exit;
  if SourceFileNameLength <= 0 then
    Exit;
  if DestinationFileNameLength <= 0 then
    Exit;
  SourceFileName := Copy(Line, SourceFileNameStart, SourceFileNameLength);
  DestinationFileName := Copy(Line, DestinationFileNameStart, DestinationFileNameLength);
  Result := True;
end;

function TFileList.DeleteTillExtension: boolean;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
  CurrentPos, ExtensionPos: integer;
begin
  Result := False;
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    Exit;
  if not TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
    Exit;
  if not DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, ExtensionPos) then
    Exit;
  CurrentPos := EditorInfo.CurPos + 1;
  if CurrentPos < DestinationFileNameStart then
    Exit;
  if CurrentPos >= ExtensionPos then
    Exit;
  Delete(Line, CurrentPos, ExtensionPos - CurrentPos);
  if TFarEditor.SetString(EditorID, EditorInfo.CurLine, Line) then
  begin
    TFarEditor.Refresh(EditorID);
    Result := True;
  end;
end;

destructor TFileList.Destroy;
begin
  if FileName <> '' then
    DeleteFile(FileName);
  inherited;
end;

class function TFileList.FindByEditorID(EditorID: TIntPtr; out Editor: TFileList): boolean;
begin
  Result := inherited FindByEditorID(EditorID, TFarEditorManaged(Editor));
end;

function TFileList.GetDescription: string;
begin
  if Options.TinyHelp then
    Result := GetMsg(MFileListDescriptionShort)
  else
    Result := GetMsg(MFileListDescription);
end;

function TFileList.GetFileAttributes(const FileName: string; out Attributes: DWORD): boolean;
var
  LastError: DWORD;
  SR: TSearchRec;
begin
  Attributes := Windows.GetFileAttributes(PChar(FileName));
  if Attributes <> INVALID_FILE_ATTRIBUTES then
  begin
    Result := True;
    Exit;
  end;
  LastError := GetLastError;
  if (LastError = ERROR_FILE_NOT_FOUND) or (LastError = ERROR_PATH_NOT_FOUND) or (LastError = ERROR_INVALID_NAME) then
  begin
    Result := False;
    Exit;
  end;
  if FindFirst(FileName, faAnyFile, SR) = 0 then
  begin
    try
      Attributes := SR.Attr;
      Result := True;
    finally
      SysUtils.FindClose(SR);
    end;
    Exit;
  end;
  Result := False;
end;

function TFileList.GetTranslatedFileName(const FileName: string): string;
begin
  Result := AnsiQuotedStr(FileName, '"');
end;

function TFileList.InternalJumpToDestinationFileExt(out Row, CurPos, LineLen: TIntPtr): TInternalJumpResult;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength: integer;
  ExtensionPos: integer;
begin
  Result := ijFailed;
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    Exit;
  if not TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
    Exit;
  if not DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, ExtensionPos) then
    Exit;
  Row := EditorInfo.CurLine;
  CurPos := ExtensionPos - 1;
  LineLen := Length(Line);
  case Options.SmartEndBehavior of
    //sebExtStartLineEnd:
    sebExtStartExtEnd:
      begin
        LineLen := DestinationFileNameStart + DestinationFileNameLength - 1;
      end;
    sebExtEndLineEnd:
      begin
        CurPos := DestinationFileNameStart + DestinationFileNameLength - 1;
      end;
  end;
  if EditorInfo.CurPos = CurPos then
  begin
    Result := ijAlreadyThere;
  end
  else if TFarEditor.SetCursorPosition(EditorID, Row, CurPos) then
  begin
    Refresh;
    Result := ijSucceeded;
  end;
end;

function TFileList.InternalJumpToDestinationFileName(out Row, CurPos: TIntPtr): TInternalJumpResult;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos: integer;
  DestinationFileNameEnd: integer;
  LeftPos: integer;
begin
  Result := ijFailed;
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    Exit;
  if not TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
    Exit;
  if not DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos) then
    Exit;
  Row := EditorInfo.CurLine;
  CurPos := DestinationFileNameStart - 1;
  if EditorInfo.CurPos = CurPos then
  begin
    Result := ijAlreadyThere;
  end
  else
  begin
    case Options.CursorMovement of
      cmoMinimizeScrolling:
        LeftPos := -1;
      cmoPrioritizeSourceFileName:
        LeftPos := 0;
      cmoPrioritizeDestinationFileName:
        begin
          DestinationFileNameEnd := DestinationFileNameStart + DestinationFileNameLength + 2; // 1 for the terminating quote, 1 for zero-based position
          if (DestinationFileNameEnd < EditorInfo.WindowSizeX) then
            LeftPos := 0
          else
            LeftPos := DestinationFileNameEnd - EditorInfo.WindowSizeX;
        end;
      cmoDestinationFileNameOnly:
        LeftPos := CurPos - 1;
      else
        LeftPos := -1;
    end;
    if TFarEditor.SetCursorPosition(EditorID, Row, CurPos, -1, -1, LeftPos, -1) then
    begin
      TFarEditor.Refresh(EditorID);
      Result := ijSucceeded;
    end;
  end;
end;

function TFileList.IsEditorLine: boolean;
var
  EditorInfo: TEditorInfo;
  Line: string;
  SourceFileName, DestinationFileName: string;
begin
  Result := False;
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    Exit;
  if not TFarEditor.GetString(EditorID, EditorInfo.CurLine, Line) then
    Exit;
  if not DecodeEditorLine(Line, SourceFileName, DestinationFileName) then
    Exit;
  Result := True;
end;

function TFileList.JumpToDestinationFileExt: boolean;
var
  Row, CurPos, LineLen: TIntPtr;
begin
  Result := InternalJumpToDestinationFileExt(Row, CurPos, LineLen) <> ijFailed;
end;

function TFileList.JumpToDestinationFileName: boolean;
var
  Row, CurPos: TIntPtr;
begin
  Result := InternalJumpToDestinationFileName(Row, CurPos) <> ijFailed;
end;

function TFileList.LoadFromPanel(AActivePanel: boolean): boolean;
var
  Panel: TFarPanel;
begin
  Panel := TFarPanel.Create(AActivePanel);
  try
    Result := LoadFromPanel(Panel);
  finally
    FreeAndNil(Panel);
  end;
end;

function TFileList.LoadFromFileList(const Directory: string; const Files: TStrings): boolean;
var
  Text: TTextStream;
  Description: string;
  i, n, MaxLength: integer;
begin
  Result := False;
  MaxLength := 0;
  for i := 0 to Pred(Files.Count) do
  begin
    n := Length(GetTranslatedFileName(Files[i]));
    if n > MaxLength then
      MaxLength := n;
  end;
  if MaxLength > 0 then
  begin
    Description := GetDescription;
    Description := StringReplace(Description, '${DIRECTORY}', Directory, [rfReplaceAll]);
    SourceDirectory := IncludeTrailingPathDelimiter(FullFileName(Directory, True));
    Text := TTextStream.Create(FileName, fmCreate);
    try
      Text.Encoding := feUtf8;
      if Description <> '' then
        Text.WriteLine(Description);
      for i := 0 to Pred(Files.Count) do
      begin
        AddFileName(Text, GetTranslatedFileName(Files[i]), MaxLength);
      end;
      Result := True;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.LoadFromPanel(const AInfo: TFarPanel): boolean;
var
  Text: TTextStream;
  Description: string;
  i, n, MaxLength: integer;
  SelectedItems: boolean;
begin
  Result := False;
  if (AInfo.Flags and OPIF_REALNAMES) = 0 then
  begin
    ShowMessage([GetMsg(MTitle), GetMsg(MUnsupportedPanel), GetMsg(MOk)], FMSG_WARNING, 1);
    Abort;
  end;
  MaxLength := 0;
  for i := 0 to Pred(AInfo.ItemCount) do
  begin
    if AInfo.Items[i].Selected then
    begin
      n := Length(GetTranslatedFileName(AInfo.Items[i].FileName));
      if n > MaxLength then
        MaxLength := n;
    end;
  end;
  SelectedItems := MaxLength > 0;
  if not SelectedItems then
  begin
    MaxLength := Length(GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName));
  end;
  if MaxLength > 0 then
  begin
    Description := GetDescription;
    Description := StringReplace(Description, '${DIRECTORY}', AInfo.Directory, [rfReplaceAll]);
    SourceDirectory := IncludeTrailingPathDelimiter(FullFileName(AInfo.Directory, True));
    Text := TTextStream.Create(FileName, fmCreate);
    try
      Text.Encoding := feUtf8;
      if Description <> '' then
        Text.WriteLine(Description);
      if SelectedItems then
      begin
        for i := 0 to Pred(AInfo.ItemCount) do
        begin
          if AInfo.Items[i].Selected then
          begin
            AddFileName(Text, GetTranslatedFileName(AInfo.Items[i].FileName), MaxLength);
          end;
        end;
      end
      else
      begin
        AddFileName(Text, GetTranslatedFileName(AInfo.Items[AInfo.CurrentItemIndex].FileName), MaxLength);
      end;
      Result := True;
    finally
      FreeAndNil(Text);
    end;
  end;
end;

function TFileList.OpenInEditor: boolean;
{$IFDEF FAR3_UP}
const
  KEY_ALT_F1: TFarKey = (VirtualKeyCode: VK_F1; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F2: TFarKey = (VirtualKeyCode: VK_F2; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F3: TFarKey = (VirtualKeyCode: VK_F3; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F4: TFarKey = (VirtualKeyCode: VK_F4; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F5: TFarKey = (VirtualKeyCode: VK_F5; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
  KEY_ALT_F6: TFarKey = (VirtualKeyCode: VK_F6; ControlKeyState: LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED);
{$ENDIF}
var
  Description: string;
  i, DescriptionLines: integer;
  Labels: TFarEditor.TEditorKeyBarLabels;
begin
  Result := False;
  if (FileName <> '') and FileExists(FileName) then
  begin
    Description := GetDescription;
    if Description = '' then
    begin
      DescriptionLines := 0;
    end
    else
    begin
      DescriptionLines := 1;
      for i := 1 to Length(Description) do
        if Description[i] = #10 then
          Inc(DescriptionLines);
    end;
    if Show(GetMsg(MTitle), 0, 0, -1, -1,
          EF_NONMODAL or EF_IMMEDIATERETURN or EF_DELETEONLYFILEONCLOSE or EF_DISABLEHISTORY {$IFDEF FAR3_UP} or EF_DISABLESAVEPOS {$ENDIF} ,
          DescriptionLines+1, 0, CP_UTF8)
    then
    begin
      if Options.SelectDestinationFilename then
        JumpToDestinationFileName;
      if not Options.DisableInternalHotkeys then
      begin
        SetLength(Labels, 6);
        {$IFDEF FAR3_UP}
        Labels[0].Key := KEY_ALT_F1;
        {$ENDIF}
        Labels[0].Caption := GetMsg(MKeyJumpToDestinationFileName);
        Labels[0].Description := GetMsg(MKeyJumpToDestinationFileNameLong);
        {$IFDEF FAR3_UP}
        Labels[1].Key := KEY_ALT_F2;
        {$ENDIF}
        Labels[1].Caption := GetMsg(MKeySelectDestinationFileNames);
        Labels[1].Description := GetMsg(MKeySelectDestinationFileNamesLong);
        {$IFDEF FAR3_UP}
        Labels[2].Key := KEY_ALT_F3;
        {$ENDIF}
        Labels[2].Caption := GetMsg(MKeyDeleteTillExtension);
        Labels[2].Description := GetMsg(MKeyDeleteTillExtensionLong);
        {$IFDEF FAR3_UP}
        Labels[3].Key := KEY_ALT_F4;
        {$ENDIF}
        Labels[3].Caption := GetMsg(MKeyJumpToDestinationFileExt);
        Labels[3].Description := GetMsg(MKeyJumpToDestinationFileExtLong);
        {$IFDEF FAR3_UP}
        Labels[4].Key := KEY_ALT_F5;
        {$ENDIF}
        Labels[4].Caption := GetMsg(MKeySmartJumpToBegin);
        Labels[4].Description := GetMsg(MKeySmartJumpToBeginLong);
        {$IFDEF FAR3_UP}
        Labels[5].Key := KEY_ALT_F6;
        {$ENDIF}
        Labels[5].Caption := GetMsg(MKeySmartJumpToEnd);
        Labels[5].Description := GetMsg(MKeySmartJumpToEndLong);
        SetKeyBar(Labels);
      end;
      Result := True;
    end
    else
      ShowWarning(GetMsg(MFailedToOpenEditor));
  end;
end;

function TFileList.ProcessKeyEvent(const Info: TKeyEventRecord): boolean;
var
  ShiftFlags: DWORD;
begin
  Result := False;
  if Options.DisableInternalHotkeys then
    Exit;
  if Info.bKeyDown then
  begin
    ShiftFlags := Info.dwControlKeyState and (LEFT_ALT_PRESSED or LEFT_CTRL_PRESSED or RIGHT_ALT_PRESSED or RIGHT_CTRL_PRESSED or SHIFT_PRESSED);
    if (ShiftFlags = LEFT_ALT_PRESSED) or (ShiftFlags = RIGHT_ALT_PRESSED) then
    begin
      case Info.wVirtualKeyCode of
        VK_F1:
          Result := JumpToDestinationFileName;
        VK_F2:
          Result := SelectDestinationFilenames;
        VK_F3:
          Result := DeleteTillExtension;
        VK_F4:
          Result := JumpToDestinationFileExt;
        VK_F5:
          Result := SmartJumpToBegin;
        VK_F6:
          Result := SmartJumpToEnd;
      end;
    end;
  end;
end;

function TFileList.RenameFiles: boolean;

  function ExtractFileNameFromFileNameList(const FileNames: string; var CurrentPos: integer; out FileName: string): boolean;
  var
    StartPos, EndPos: integer;
  begin
    StartPos := CurrentPos;
    EndPos := Length(FileNames);
    while CurrentPos <= EndPos do
    begin
      if (FileNames[CurrentPos] = ',') or (FileNames[CurrentPos] = ';') then
      begin
        FileName := Trim(Copy(FileNames, StartPos, CurrentPos - StartPos));
        Inc(CurrentPos);
        Result := True;
        Exit;
      end;
      Inc(CurrentPos);
    end;
    FileName := Trim(Copy(FileNames, StartPos, CurrentPos - StartPos));
    Result := (FileName <> '');
  end;

var
  Text: TTextStream;
  FailedText: TTextStream;
  FailedStream: TMemoryStream;
  Line, SourceFileName, DestinationFileName, SrcFN, DestFN: string;
  SrcAttr, DestAttr: DWORD;
  Again, AlwaysSkip: boolean;
  Failed: boolean;
  Panel: TFarPanel;
  i, j, FileIndex: Integer;
  DescriptionFileNames: TStringList;
  Descriptions: array of TFarDescriptions;
  SkipDescriptionErrors: boolean;
  DescriptionFileNamesExt, DescriptionFileName: string;
  DescriptionFileNamesPos: integer;
begin
  Result := False;
  if (FileName <> '') and FileExists(FileName) then
  begin
    Panel := nil;
    FailedText := nil;
    DescriptionFileNames := nil;
    SetLength(Descriptions, 0);
    Text := TTextStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      FailedStream := TMemoryStream.Create;
      FailedText := TTextStream.Create(FailedStream, True);
      AlwaysSkip := False;
      Result := True;
      if Options.SelectRenamedFiles then
      begin
        Panel := TFarPanel.Create(True, False);
        if AnsiCompareText(IncludeTrailingPathDelimiter(FullFileName(Panel.Directory, True)), SourceDirectory) <> 0 then
          FreeAndNil(Panel);
        if Panel <> nil then
        begin
          Panel.Refresh;
          if not RepeatedRun then
          begin
            for i := 0 to Pred(Panel.ItemCount) do
              Panel.Items[i].Selected := False;
            Panel.ApplySelection;
          end;
        end;
      end;
      if Options.UpdateDescriptions then
      begin
        DescriptionFileNames := TFarDescriptions.GetDescriptionFileNames;
        DescriptionFileNamesExt := Options.DescriptionFiles;
        DescriptionFileNamesPos := 1;
        while ExtractFileNameFromFileNameList(DescriptionFileNamesExt, DescriptionFileNamesPos, DescriptionFileName) do
        begin
          for i := 0 to Pred(DescriptionFileNames.Count) do
          begin
            if SameText(DescriptionFileNames[i], DescriptionFileName) then
            begin
              DescriptionFileName := '';
              Break;
            end;
          end;
          if DescriptionFileName <> '' then
            DescriptionFileNames.Add(DescriptionFileName);
        end;
        SetLength(Descriptions, DescriptionFileNames.Count);
        for i := 0 to Pred(Length(Descriptions)) do
          Descriptions[i] := nil;
        j := 0;
        for i := 0 to Pred(Length(Descriptions)) do
        begin
          if DescriptionFileNames[i] <> '' then
          begin
            if FileExists(SourceDirectory + DescriptionFileNames[i]) then
            begin
              Descriptions[j] := TFarDescriptions.Create;
              try
                Descriptions[j].Load(SourceDirectory + DescriptionFileNames[i]);
                Inc(j);
              except
                FreeAndNil(Descriptions[j]);
              end;
            end;
          end;
        end;
        SetLength(Descriptions, j);
      end;
      SkipDescriptionErrors := False;
      while Text.ReadLine(Line) do
      begin
        Failed := True;
        if DecodeEditorLine(Line, SourceFileName, DestinationFileName) then
        begin
          repeat
            Again := False;
            if IsRelativePath(SourceFileName) then
              SrcFN := SourceDirectory + SourceFileName
            else
              SrcFN := SourceFileName;
            if IsRelativePath(DestinationFileName) then
              DestFN := SourceDirectory + DestinationFileName
            else
              DestFN := DestinationFileName;
            if SrcFN = DestFN then
            begin
              Failed := False;
              Break;
            end
            else if not GetFileAttributes(SrcFN, SrcAttr) then
            begin
              ShowMessage([GetMsg(MTitle), GetMsg(MSourceFileNotFound), SrcFN, GetMsg(MOk)], FMSG_WARNING, 1);
            end
            else if GetFileAttributes(DestFN, DestAttr) and ((DestAttr and FILE_ATTRIBUTE_DIRECTORY) <> 0) and (AnsiCompareText(SrcFN, DestFN) <> 0) then
            begin
              ShowMessage([GetMsg(MTitle), GetMsg(MCantOverwriteDirectory), DestFN, GetMsg(MOk)], FMSG_WARNING, 1);
              Result := False;
            end
            else if not RenameFile(SrcFN, DestFN) then
            begin
              if AlwaysSkip then
                Result := False
              else
              begin
                case ShowMessage([GetMsg(MTitle), GetMsg(MFailedToRename), SrcFN, GetMsg(MFailedToRenameTo), DestFN, GetMsg(MRetry), GetMsg(MOverwrite), GetMsg(MSkip), GetMsg(MSkipAll)], FMSG_WARNING, 4) of
                  0:   Again := True;
                  1:   begin
                         DeleteFile(DestFN);
                         Again := True;
                       end;
                  2:   Result := False;
                  3:   begin
                         AlwaysSkip := True;
                         Result := False;
                       end;
                  else Abort;
                end;
              end
            end
            else
            begin
              Failed := False;
              for i := 0 to Pred(Length(Descriptions)) do
              begin
                try
                  Descriptions[i].Items.Rename(SourceFileName, DestinationFileName);
                except
                  on E: Exception do
                    if not SkipDescriptionErrors then
                      if ShowMessage([GetMsg(MTitle), GetMsg(MFailedToUpdateFileDescription), Format(GetMsg(MFailedToUpdateFileDescriptionFile), [Descriptions[i].FileName]), Format(GetMsg(MFailedToUpdateFileDescriptionDescription), [SourceFileName]), Format(GetMsg(MFailedToUpdateFileDescriptionError), [E.ClassName, E.Message]), GetMsg(MOk), GetMsg(MSkipAll)], FMSG_WARNING, 2) = 2 then
                        SkipDescriptionErrors := True;
                end;
              end;
            end;
          until not Again;
        end;
        if Failed then
        begin
          FailedText.WriteLine(Line);
        end
        else if Panel <> nil then
        begin
          Panel.Refresh;
          if Panel.FindFile(DestinationFileName, FileIndex) then
          begin
            Panel.Items[FileIndex].Selected := True;
            Panel.ApplySelection;
          end;
        end;
      end;
      if not Result then
      begin
        FreeAndNil(Text);
        FailedText.Flush;
        FailedStream.SaveToFile(FileName);
      end;
    finally
      FreeAndNil(DescriptionFileNames);
      for i := 0 to Pred(Length(Descriptions)) do
      begin
        Descriptions[i].Save;
        FreeAndNil(Descriptions[i]);
      end;
      SetLength(Descriptions, 0);
      FreeAndNil(Text);
      FreeAndNil(FailedText);
      FreeAndNil(Panel);
      TFarPanel.Refresh(True);
    end;
  end;
end;

function TFileList.SelectDestinationFilenames: boolean;
var
  EditorInfo: TEditorInfo;
  Select: TEditorSelect;
  LineNr, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos: integer;
  Line: string;
begin
  Result := False;
  if TFarEditor.GetInfo(EditorID, EditorInfo) then
  begin
    FillChar(Select, Sizeof(Select), 0);
    {$IFDEF FAR3_UP}
    Select.StructSize := Sizeof(Select);
    {$ENDIF}
    Select.BlockType := BTYPE_COLUMN;
    Select.BlockStartLine := -1;
    Select.BlockStartPos := -1;
    Select.BlockWidth := -1;
    Select.BlockHeight := -1;
    for LineNr := 0 to Pred(EditorInfo.TotalLines) do
    begin
      if TFarEditor.GetString(EditorID, LineNr, Line) then
      begin
        if DecodeEditorLine(Line, SourceFileNameStart, SourceFileNameLength, DestinationFileNameStart, DestinationFileNameLength, DestinationExtensionPos) then
        begin
          if Select.BlockStartLine < 0 then
            Select.BlockStartLine := LineNr;
          if (Select.BlockStartPos < 0) or (Select.BlockStartPos > (DestinationFileNameStart-2)) then
            Select.BlockStartPos := DestinationFileNameStart-2;
          if Select.BlockWidth < (DestinationFileNameLength+2) then
            Select.BlockWidth := DestinationFileNameLength+2;
          Select.BlockHeight := LineNr - Select.BlockStartLine + 1;
          Result := True;
        end;
      end;
    end;
    if Result then
    begin
      Result := FarApi.EditorControl( {$IFDEF FAR3_UP} EditorID, {$ENDIF} ECTL_SELECT, {$IFDEF FAR3_UP} 0, {$ENDIF} @Select) <> 0;
      TFarEditor.Refresh(EditorID);
    end;
  end;
end;

function TFileList.ShowEditorMenu: boolean;
var
  SelectedItem: integer;
begin
  Result := False;
  if ShowMenu(MENU_EDITOR_GUID, GetMsg(MTitle), '', [
          GetMsg(MKeyJumpToDestinationFileNameLong)
        , GetMsg(MKeySelectDestinationFileNamesLong)
        , GetMsg(MKeyDeleteTillExtensionLong)
        , GetMsg(MKeyJumpToDestinationFileExtLong)
        , GetMsg(MKeySmartJumpToBeginLong)
        , GetMsg(MKeySmartJumpToEndLong)
      ], SelectedItem) then
  begin
    case SelectedItem of
      0:   Result := JumpToDestinationFileName;
      1:   Result := SelectDestinationFilenames;
      2:   Result := DeleteTillExtension;
      3:   Result := JumpToDestinationFileExt;
      4:   Result := SmartJumpToBegin;
      5:   Result := SmartJumpToEnd;
    end;
  end;
end;

function TFileList.SmartJumpToBegin: boolean;
var
  Row, CurPos: TIntPtr;
begin
  Result := False;
  case InternalJumpToDestinationFileName(Row, CurPos) of
    ijSucceeded:
      Result := True;
    ijAlreadyThere:
      begin
        if TFarEditor.SetCursorPosition(EditorID, Row, 0) then
        begin
          TFarEditor.Refresh(EditorID);
          Result := True;
        end;
      end;
  end;
end;

function TFileList.SmartJumpToEnd: boolean;
var
  Row, CurPos, LineLen: TIntPtr;
begin
  Result := False;
  case InternalJumpToDestinationFileExt(Row, CurPos, LineLen) of
    ijSucceeded:
      Result := True;
    ijAlreadyThere:
      begin
        if TFarEditor.SetCursorPosition(EditorID, Row, LineLen) then
        begin
          TFarEditor.Refresh(EditorID);
          Result := True;
        end;
      end;
  end;
end;

end.

