unit uPlugin;
{$INCLUDE 'Plugin.inc'}

interface

implementation

uses
  Windows, SysUtils, Classes, Variants,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  {$IFDEF DEBUG_LOGS} uLog, {$ENDIF}
  uFileList,
  uOptions,
  uOptionsUI,
  uMessages,
  uPluginConsts;

type
  TPlugin = class(TFarPlugin)
  private
    fOptions: TOptions;
    function GetOptions: TOptions;
  protected
    function PluginID: TGUID; override;
    function GetPluginTitle(out Description, Author: string): string; override;
    procedure GetPluginInfo(var Flags: TPluginFlags; var MainMenu, ConfigMenu, DiskMenu: TFarPlugin.TFarPluginMenu; var CommandPrefix: string); override;
    function GetCommonString(ID: TFarPlugin.TCommonString): string; override;
    function PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean; override;
    procedure DoLog(const Msg: string); override;
    procedure RefreshOptions;
    property Options: TOptions read GetOptions;
    function Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle; override;
    function ExecuteFromPanel: boolean;
    function ExecuteFromFileList(const Directory: string; const Files: TStrings): boolean;
    function ProcessMacroCall(const Params: array of Variant): boolean;
  public
    constructor Create;
    destructor Destroy; override;
  end;

const
  PLUGIN_TITLE = 'FRename3';
  PLUGIN_DESCRIPTION = 'Files Rename 3';
  PLUGIN_AUTHOR = 'Pepak, https://www.pepak.net';

{ TPlugin }

constructor TPlugin.Create;
begin
  inherited;
  fOptions := nil;
end;

destructor TPlugin.Destroy;
begin
  FreeAndNil(fOptions);
  inherited;
end;

procedure TPlugin.DoLog(const Msg: string);
begin
  {$IFDEF DEBUG_LOGS}
  uLog.Log(Msg);
  {$ENDIF}
end;

function TPlugin.Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle;
var
  List: TFileList;
begin
  Result := 0;
  RefreshOptions;
  if OpenFrom = OPEN_FROMMACRO_ then
  begin
    if ProcessMacroCall(TFarMacro.MacroParamsToDelphi(Data)) then
      Result := 1;
    Exit;
  end;
  if OpenFrom = OPEN_EDITOR then
  begin
    if TFileList.FindByEditorID(-1, List) then
      List.ShowEditorMenu;
  end
  else
  begin
    if not ExecuteFromPanel then
      ShowWarning(GetMsg(MFailedToLoadFileList));
  end;
end;

function TPlugin.ExecuteFromFileList(const Directory: string; const Files: TStrings): boolean;
var
  List: TFileList;
begin
  Result := False;
  List := TFileList.Create(Options);
  try
    if not List.LoadFromFileList(Directory, Files) then
      Exit;
    if not List.OpenInEditor then
      Exit;
    List := nil; // This prevents the editor from being destroyed
    Result := True;
  finally
    FreeAndNil(List);
  end;
end;

function TPlugin.ExecuteFromPanel: boolean;
var
  List: TFileList;
begin
  Result := False;
  List := TFileList.Create(Options);
  try
    if not List.LoadFromPanel(True) then
      Exit;
    if not List.OpenInEditor then
      Exit;
    List := nil; // This prevents the editor from being destroyed
    Result := True;
  finally
    FreeAndNil(List);
  end;
end;

function TPlugin.GetCommonString(ID: TFarPlugin.TCommonString): string;
begin
  case ID of
    TFarPlugin.TCommonString.MTitle:
      Result := GetMsg(MTitle);
    TFarPlugin.TCommonString.MExceptionOccurredIn:
      Result := GetMsg(MExceptionOccurredIn);
    TFarPlugin.TCommonString.MOk:
      Result := GetMsg(MOk);
    TFarPlugin.TCommonString.MCancel:
      Result := GetMsg(MCancel);
    TFarPlugin.TCommonString.MYes:
      Result := GetMsg(MYes);
    TFarPlugin.TCommonString.MNo:
      Result := GetMsg(MNo);
    else
      Result := inherited GetCommonString(ID);
  end;
end;

function TPlugin.GetOptions: TOptions;
begin
  if fOptions = nil then
  begin
    fOptions := TOptions.Create;
    fOptions.Load;
  end;
  Result := fOptions;
end;

procedure TPlugin.GetPluginInfo(var Flags: TPluginFlags; var MainMenu, ConfigMenu, DiskMenu: TFarPlugin.TFarPluginMenu; var CommandPrefix: string);
begin
  Flags := Flags or PF_EDITOR;
  SetLength(MainMenu, 1);
  MainMenu[0].Title := GetMsg(MTitle);
  MainMenu[0].ID := MENU_PANEL_GUID;
  SetLength(ConfigMenu, 1);
  ConfigMenu[0].Title := GetMsg(MTitle);
  ConfigMenu[0].ID := MENU_OPTIONS_GUID;
end;

function TPlugin.GetPluginTitle(out Description, Author: string): string;
begin
  Result := PLUGIN_TITLE;
  Description := PLUGIN_DESCRIPTION;
  Author := PLUGIN_AUTHOR;
end;

function TPlugin.PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean;
var
  Dialog: TOptionsDialog;
begin
  Result := False;
  RefreshOptions;
  Dialog := TOptionsDialog.Create;
  try
    Dialog.Options := Options;
    if Dialog.Execute then
    begin
      Options.Save;
      Result := True;
    end;
  finally
    FreeAndNil(Dialog);
  end;
end;

function TPlugin.PluginID: TGUID;
begin
  Result := PLUGIN_GUID;
end;

function TPlugin.ProcessMacroCall(const Params: array of Variant): boolean;
var
  Command: string;
  List: TStringList;
  Editor: TFileList;
begin
  Result := False;
  if Length(Params) > 0 then
  begin
    Command := Params[0];
    if Command = 'panel' then
    begin
      Result := ExecuteFromPanel;
      Exit;
    end;
    if Command = 'files' then
    begin
      if Length(Params) <> 3 then
        raise EFarPluginError.Create(GetMsg(MInvalidMacroParameter));
      List := TFarMacro.LuaParamToStringList(Params[2]);
      try
        Result := ExecuteFromFileList(Params[1], List);
      finally
        FreeAndNil(List);
      end;
      Exit;
    end;
    if Command = 'is_editor' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        Result := True;
      Exit;
    end;
    if Command = 'is_editor_line' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.IsEditorLine then
          Result := True;
      Exit;
    end;
    if Command = 'cursor_to_filename' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.JumpToDestinationFileName then
          Result := True;
      Exit;
    end;
    if Command = 'cursor_to_extension' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.JumpToDestinationFileExt then
          Result := True;
      Exit;
    end;
    if Command = 'smart_home' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.SmartJumpToBegin then
          Result := True;
      Exit;
    end;
    if Command = 'smart_end' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.SmartJumpToEnd then
          Result := True;
      Exit;
    end;
    if Command = 'select_filenames' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.SelectDestinationFilenames then
          Result := True;
      Exit;
    end;
    if Command = 'delete_till_extension' then
    begin
      if TFileList.FindByEditorID(-1, Editor) then
        if Editor.DeleteTillExtension then
          Result := True;
    end;
  end;
end;

procedure TPlugin.RefreshOptions;
begin
  FreeAndNil(fOptions);
end;

exports
  ProcessEditorEventW,
  ProcessEditorInputW;

initialization
  TFarPlugin.InitInstance(TPlugin.Create);

end.
